/*! 
    \file animation3d.h
*/

#include "animation.h"

#include "vector.h"
class PNode;

//! KeyFrame3D Struct
/**
   This represents one point with direction of the animation
*/
typedef struct KeyFrame3D {
  float duration;
  Vector position;
  Quaternion direction;
  ANIM_FUNCTION animFunc;
};

//! Animation Struct
/**
   This represents an animation for a object
*/
class Animation3D : public Animation
{
 public:
  Animation3D(PNode* object);
  virtual ~Animation3D(void);
    
  virtual void rewind(void);

  void addKeyFrame(Vector position, Quaternion direction, float time, ANIM_FUNCTION animFunc = ANIM_LINEAR);
  void addKeyFrame(KeyFrame3D* frame);

  virtual void tick(float dt);

  // animation functions
  void setAnimFunc(ANIM_FUNCTION animFunc);

 private:
  void constant(float timePassed) const;
  void linear(float timePassed) const;
  void sine(float timePassed) const;
  void cosine(float timePassed) const;
  void exp(float timePassed) const;
  void negExp(float timePassed) const;
  void quadratic(float timePassed) const;
  void random(float timePassed) const;
  //  ANIM_FUNCTION animFunc;
  KeyFrame3D* currentKeyFrame;
  KeyFrame3D* nextKeyFrame;
  tList<KeyFrame3D>* keyFrameList;
  void (Animation3D::*animFunc)(float) const;

  // more class-local description
  PNode* object;
  Vector lastPosition;
  Vector tmpVect;
  float deltaT;
};
